// ---------------------------------------------------------------
// Digger for Smartphone by Lutz Roeder, June 2000, November 2003.
// Graphics and Levels by Alexander Lang and Jan Fricke.
// http://www.aisto.com/roeder
// ---------------------------------------------------------------
namespace Digger
{
	using System;
	using System.Collections;
	using System.Drawing;
	using System.IO;
	using System.Windows.Forms;

	internal class ApplicationWindow : Form
	{
		private GameWindow window = new GameWindow();

		public ApplicationWindow()
		{
			this.Icon = new Icon(this.GetType().Assembly.GetManifestResourceStream("Digger.Application.ico"));
			this.Text = "Digger";
			this.MaximizeBox = false;
			this.ClientSize = this.window.ClientSize;
			this.Controls.Add(this.window);
		}

		protected override void OnGotFocus(EventArgs e)
		{
			base.OnGotFocus(e);
			this.window.Focus();
		}

		private static void Main(string[] args)
		{
			Application.Run(new ApplicationWindow());		
		}
	}

	internal class GameWindow : Panel
	{ 
		private byte[][] lastScreen = null;
		private Bitmap[] sprites = new Bitmap[34];
		private Bitmap[] characters = new Bitmap[63];
		private Position origin = new Position(0, 0);
		private Timer timer;
		private int tick = 0;
		private Level level = new Level();
		private int lives = 20;
		private bool[] keysRelease = new bool[4];
		private bool[] keysDelay = new bool[4];  
  
		public GameWindow()
		{
			this.ClientSize = new Size(176, 200);

			using (Stream stream = this.GetType().Assembly.GetManifestResourceStream("Digger.Sprite.list"))
			{
				BinaryReader reader = new BinaryReader(stream);
				for (int i = 0; i < 34; i++)
				{
					int length = reader.ReadInt32();
					byte[] bytes = reader.ReadBytes(length);
					MemoryStream memoryStream = new MemoryStream(bytes);
					this.sprites[i] = new Bitmap(memoryStream);
					memoryStream.Close();
				}
			}

			using (Stream stream = this.GetType().Assembly.GetManifestResourceStream("Digger.Font.list"))
			{
				BinaryReader reader = new BinaryReader(stream);
				for (int i = 0; i < 63; i++)
				{
					int length = reader.ReadInt32();
					byte[] bytes = reader.ReadBytes(length);
					MemoryStream memoryStream = new MemoryStream(bytes);
					this.characters[i] = new Bitmap(memoryStream);
					memoryStream.Close();
				}
			}

			this.lives = 20;
			this.level.Number = 0;
			this.level.Load();

			this.timer = new Timer();
			this.timer.Interval = 40;
			this.timer.Tick += new EventHandler(this.Timer_Tick);
			this.timer.Enabled = true;

			this.lastScreen = new byte[11][];
			for (int i = 0; i < 11; i++)
			{
				this.lastScreen[i] = new byte[12];
				for (int j = 0; j < 12; j++)
				{
					this.lastScreen[i][j] = 0xff;
				}
			}
		}

		protected override void OnPaintBackground(PaintEventArgs e) 
		{
		}

		protected override void OnPaint(PaintEventArgs e)
		{
			if ((e.ClipRectangle.Width != 0) && (e.ClipRectangle.Height != 0))
			{
				string text = " [" + (this.level.Number + 1).ToString("D2") + " \\" + this.lives.ToString("D2") + " ]" + level.Collected.ToString("D2") + "/" + this.level.Diamonds.ToString("D2") + " ^" + this.level.Time.ToString("D4") + " ";
				Point position = new Point(0, 0);
				for (int i = 0; i < text.Length; i++) 
				{
					e.Graphics.DrawImage(this.characters[(byte) (text[i] - 32)], position.X, position.Y);
					position.X += 8;
				}

				for (int x = 0; x < 11; x++)
				{
					for (int y = 0; y < 12; y++)
					{
						Rectangle rectangle = new Rectangle(x * 16, 8 + (y * 16), 16, 16);
						if (e.ClipRectangle.IntersectsWith(rectangle)) 
						{
							Sprite sprite = level.GetSprite(this.origin.X + x, this.origin.Y + y);
							byte imageIndex = sprite.ImageIndex;
							if (imageIndex != this.lastScreen[x][y])
							{
								e.Graphics.DrawImage(this.sprites[imageIndex], rectangle.X, rectangle.Y);
								this.lastScreen[x][y] = imageIndex;
							}
						}
					}
				}
			}
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			switch (e.KeyCode)
			{
				case Keys.Left:
					level.Digger.Keys[0] = true; 
					keysDelay[0] = false; 
					break;

				case Keys.Right:
					level.Digger.Keys[1] = true; 
					keysDelay[1] = false; 
					break;

				case Keys.Up:
					level.Digger.Keys[2] = true; 
					keysDelay[2] = false; 
					break;

				case Keys.Down:
					level.Digger.Keys[3] = true; 
					keysDelay[3] = false; 
					break;

				case Keys.F8:
					if (this.lives > 0)
					{
						this.lives--;
						this.level.Load();
					}
					else
					{
						// Restart game
						this.lives = 20;
						this.level.Number = 0;
						this.level.Load();
					}
					break;

				case Keys.F9:
					ApplicationWindow applicationWindow = (ApplicationWindow) this.TopLevelControl;
					applicationWindow.Close();
					break;

				case Keys.D0:
					if (this.level.Number < 29)
					{
						this.level.Number++;
						this.level.Load();
					}
					break;
			}
		}

		protected override void OnKeyUp(KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Left) 
			{
				if (keysDelay[0]) 
				{
					level.Digger.Keys[0] = false; 
				}
				else 
				{
					keysRelease[0] = true;
				}
			}

			if (e.KeyCode == Keys.Right) 
			{
				if (keysDelay[1]) 
				{
					level.Digger.Keys[1] = false; 
				}
				else 
				{
					keysRelease[1] = true;
				}
			}

			if (e.KeyCode == Keys.Up) 
			{
				if (keysDelay[2]) 
				{
					level.Digger.Keys[2] = false; 
				}
				else 
				{
					keysRelease[2] = true;
				}
			}

			if (e.KeyCode == Keys.Down) 
			{
				if (keysDelay[3]) 
				{
					level.Digger.Keys[3] = false; 
				}
				else 
				{
					keysRelease[3] = true;
				}
			}
		}

		private void Timer_Tick(object sender, EventArgs e)
		{
			this.tick++;
			this.level.Tick1();
			if ((this.tick % 2) == 0)
			{
				if (this.keysRelease[0]) 
				{ 
					this.level.Digger.Keys[0] = this.keysRelease[0] = false; 
				}
				
				if (this.keysRelease[1]) 
				{ 
					this.level.Digger.Keys[1] = this.keysRelease[1] = false; 
				}
				
				if (this.keysRelease[2]) 
				{ 
					this.level.Digger.Keys[2] = this.keysRelease[2] = false; 
				}
				
				if (this.keysRelease[3]) 
				{ 
					this.level.Digger.Keys[3] = this.keysRelease[3] = false; 
				}
				
				this.level.Tick4();
			}

			Position position = this.level.GetSpritePosition(this.level.Digger);

			if (this.origin.X < (20 - 11))
			{
				if (((this.origin.X + 11) - position.X) < 6)
				{
					this.origin.X++;
				}
			}

			if (this.origin.Y < (14 - 12))
			{
				if (((this.origin.Y + 12) - position.Y) < 5)
				{
					this.origin.Y++;
				}
			}

			if (this.origin.X > 0)
			{
				if ((position.X - this.origin.X) < 6)
				{
					this.origin.X--;
				}
			}

			if (this.origin.Y > 0)
			{
				if ((position.Y - this.origin.Y) < 5)
				{
					this.origin.Y--;
				}
			}

			this.Invalidate();
			this.Update();
		}

		private enum Directions 
		{ 
			Left, 
			Right, 
			Up, 
			Down, 
			None
		}
  
		private struct Position
		{
			public int X;
			public int Y;
  
			public Position(int x, int y) 
			{ 
				this.X = x; 
				this.Y = y; 
			}
			
			public Position(Position position) 
			{ 
				this.X = position.X; 
				this.Y = position.Y; 
			}

			public override bool Equals(object obj)
			{
				Position position = (Position) obj;
				return ((this.X == position.X) && (this.Y == position.Y));
			}

			public override int GetHashCode()
			{
				return base.GetHashCode();
			}

			public static bool operator==(Position a, Position b)
			{
				return a.Equals(b);
			}

			public static bool operator!=(Position a, Position b)
			{
				return !a.Equals(b);
			}

			public override string ToString()
			{
				return "(" + this.X + ", " + this.Y + ")";
			}

		}
  
		private class Sprite 
		{
			public virtual byte ImageIndex 
			{ 
				get
				{
					return 0; 
				}
			}
		}
  
		private class Ground : Sprite
		{ 
			public override byte ImageIndex
			{ 
				get
				{
					return 2; 
				}
			}
		}
  
		private class Ghost : Sprite 
		{
			public Directions Direction = Directions.None;
			public bool IsAlive = true;
  
			public override byte ImageIndex
			{
				get
				{
					switch (this.Direction)
					{
						case Directions.Left:
							return 4;
	
						case Directions.Right:
							return 5;
	
						case Directions.Up:
							return 6;
	
						case Directions.Down:
							return 3;
					}
	
					return 3;
				}
			}
  
			public virtual void Move(Level level) 
			{ 
			}
  
			public void Die()
			{
				this.IsAlive = false;
			}
  
			public void Blast(Level level)
			{
				if (this.IsAlive)
				{
					Position p = level.GetSpritePosition(this);
	  				for (int y = p.Y - 1; y <= p.Y + 1; y++)
					{
						for (int x = p.X - 1; x <= p.X + 1; x++)
						{
							if ((x > 0) && (x < 19) && (y > 0) && (y < 13))
							{
								if (level.GetSprite(x, y).GetType() == typeof(Digger))
								{
									Digger d = (Digger) level.GetSprite(x, y);
									level.Digger.Die();
								}
								else
								{
									if (level.GetSprite(x, y).GetType().BaseType == typeof(Ghost))
									{
										Ghost g = (Ghost) level.GetSprite(x, y);
										g.Die();
										level.Score += 99;
									}
	  
									level.SetSprite(x, y, new Nothing());
								}
							}
						}
					}
	  
					this.Die();
				}
			}
		}
  
		private class Ghost180 : Ghost
		{
			public override void Move(Level l)
			{
				if (this.IsAlive)
				{  
					Position p = l.GetSpritePosition(this);
					Position[] w = { new Position(p), new Position(p), new Position(p), new Position(p) };
	  
					if (this.Direction == Directions.Left)  
					{ 
						w[0].X--; 
						w[1].X++; 
					}
					
					if (this.Direction == Directions.Right) 
					{ 
						w[0].X++; 
						w[1].X--; 
					}
					
					if (this.Direction == Directions.Up)    
					{ 
						w[0].Y--; 
						w[1].Y++; 
					}
					
					if (this.Direction == Directions.Down)  
					{ 
						w[0].Y++; 
						w[1].Y--; 
					}
	  
					for (int i = 0; i < 4; i++)
					{
						if (p != w[i])
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Digger)) 
							{
								l.Digger.Die();
							}
	  
							// Nothing
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Nothing))
							{
								if (d.X < p.X) 
								{
									this.Direction = Directions.Left;
								}

								if (d.X > p.X) 
								{
									this.Direction = Directions.Right;
								}

								if (d.Y < p.Y) 
								{
									this.Direction = Directions.Up;
								}

								if (d.Y > p.Y) 
								{
									this.Direction = Directions.Down;
								}
	          
								l.SetSprite(d.X, d.Y, this);
								l.SetSprite(p.X, p.Y, new Nothing());
	  
								return;
							}
						}
					}
				}
			}
		}
  
		private class Ghost90L : Ghost
		{
			public override void Move(Level l)
			{
				if (this.IsAlive) 
				{
					Position p = l.GetSpritePosition(this);
					Position[] w = { new Position(p), new Position(p), new Position(p), new Position(p) };
	  
					if (Direction == Directions.Left)  { w[0].X--; w[1].Y++; w[2].Y--; w[3].X++; }
					if (Direction == Directions.Right) { w[0].X++; w[1].Y--; w[2].Y++; w[3].X--; }
					if (Direction == Directions.Up)    { w[0].Y--; w[1].X--; w[2].X++; w[3].Y++; }
					if (Direction == Directions.Down)  { w[0].Y++; w[1].X++; w[2].X--; w[3].Y--; }
	  
					for (int i = 0; i < 4; i++)
					{
						if (p != w[i])
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Digger)) 
							{
								l.Digger.Die();
							}
	  
							// Nothing
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Nothing))
							{
								if (d.X < p.X) Direction = Directions.Left;
								if (d.X > p.X) Direction = Directions.Right;
								if (d.Y < p.Y) Direction = Directions.Up;
								if (d.Y > p.Y) Direction = Directions.Down;
								l.SetSprite(d.X, d.Y, this);
								l.SetSprite(p.X, p.Y, new Nothing());
								return;
							}
						}
					}
				}
			}
		}
  
		private class Ghost90LR : Ghost
		{
			public Directions Lastturn = Directions.None;
  
			public override void Move(Level l)
			{
				if (this.IsAlive)
				{
					Position p = l.GetSpritePosition(this);
					
					Position[] w = 
					{ 
						new Position(p), 
						new Position(p), 
						new Position(p), 
						new Position(p) 
					};
	  
					if (Direction == Directions.Left)
					{
						w[0].X--; w[3].X++;
						if (Lastturn == Directions.Left) 
						{ 
							w[1].Y--; 
							w[2].Y++; 
						} 
						else 
						{ 
							w[1].Y++; 
							w[2].Y--; 
						};
					}
	  
					if (Direction == Directions.Right)
					{
						w[0].X++; w[3].X--;
						if (Lastturn == Directions.Left) 
						{ 
							w[1].Y++; 
							w[2].Y--; 
						}
 						else 
						{ 
							w[1].Y--; 
							w[2].Y++; 
						};
					}
	  
					if (Direction == Directions.Up)
					{
						w[0].Y--; w[3].Y++;
						if (Lastturn == Directions.Left) { w[1].X++; w[2].X--; } 
						else { w[1].X--; w[2].X++; };
					}
	  
					if (Direction == Directions.Down)
					{
						w[0].Y++; w[3].Y--;
						if (Lastturn == Directions.Left) { w[1].X--; w[2].X++; } 
						else { w[1].X++; w[2].X--; };
					}
	     
					for (int i = 0; i < 4; i++)
					{
						if (p != w[i])
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Digger)) l.Digger.Die();
	  
							// Nothing
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Nothing))
							{
								Directions LastDirection = Direction;
	  
								if (d.X < p.X) Direction = Directions.Left;
								if (d.X > p.X) Direction = Directions.Right;
								if (d.Y < p.Y) Direction = Directions.Up;
								if (d.Y > p.Y) Direction = Directions.Down;
	  
								switch (LastDirection)
								{
									case Directions.Left:
										if (Direction == Directions.Down)  Lastturn = Directions.Left; 
										if (Direction == Directions.Up)    Lastturn = Directions.Right;
										break;
									case Directions.Right:
										if (Direction == Directions.Down)  Lastturn = Directions.Right; 
										if (Direction == Directions.Up)    Lastturn = Directions.Left;
										break;
									case Directions.Up:
										if (Direction == Directions.Left)  Lastturn = Directions.Left; 
										if (Direction == Directions.Right) Lastturn = Directions.Right;
										break;
									case Directions.Down:
										if (Direction == Directions.Left)  Lastturn = Directions.Right; 
										if (Direction == Directions.Right) Lastturn = Directions.Left;
										break;
								}        
	  
								l.SetSprite(d.X, d.Y, this);
								l.SetSprite(p.X, p.Y, new Nothing());
	  
								return;
							}
						}
					}
				}
			}
		}
  
		private class Ghost90R : Ghost
		{
			public override void Move(Level l)
			{
				if (this.IsAlive)
				{
					Position p = l.GetSpritePosition(this);
					Position[] w = { new Position(p), new Position(p), new Position(p), new Position(p) };
	  
					if (Direction == Directions.Left)  { w[0].X--; w[1].Y--; w[2].Y++; w[3].X++; }
					if (Direction == Directions.Right) { w[0].X++; w[1].Y++; w[2].Y--; w[3].X--; }
					if (Direction == Directions.Up)    { w[0].Y--; w[1].X++; w[2].X--; w[3].Y++; }
					if (Direction == Directions.Down)  { w[0].Y++; w[1].X--; w[2].X++; w[3].Y--; }
	  
					for (int i = 0; i < 4; i++)
					{
						if (p != w[i])
						{
							Position d = new Position(w[i]);
	        
							// Digger
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Digger))
								l.Digger.Die();
	  
							// Nothing
							if (l.GetSprite(d.X, d.Y).GetType() == typeof(Nothing))
							{
								if (d.X < p.X) Direction = Directions.Left;
								if (d.X > p.X) Direction = Directions.Right;
								if (d.Y < p.Y) Direction = Directions.Up;
								if (d.Y > p.Y) Direction = Directions.Down;
	          
								l.SetSprite(d.X, d.Y, this);
								l.SetSprite(p.X, p.Y, new Nothing());
	  
								return;
							}
						}
					}
				}
			}
		}
  
		private class Wall : Sprite 
		{
			public override byte ImageIndex
			{ 
				get
				{
					return 14; 
				}
			}
		}
  
		private class Digger : Sprite 
		{
			public bool[] Keys = new bool[4];

			private Directions direction = Directions.None;
			private int step = 0;
			private bool isAlive = true;
			private bool[] stone = new bool[2];
  
			public override byte ImageIndex
			{
				get
				{
					if (this.isAlive) 
					{
						if ((direction == Directions.Left) && (step <= 5)) 
						{
							byte[] left = { 16, 17, 18, 19, 18, 17 };
							return left[step];
						}

						if ((direction == Directions.Right) && (step <= 5)) 
						{
							byte[] right = { 20, 21, 22, 23, 22, 21 };
							return right[step];
						}

						if ((direction == Directions.Up) && (step <= 1)) 
						{
							byte[] up = { 24, 25 };
							return up[step];
						}

						if ((direction == Directions.Down) && (step <= 1)) 
						{
							byte[] down = { 26, 27 };
							return down[step];
						}
						
						byte[] center = { 15, 15, 15, 15, 15, 15, 15, 15, 28, 28, 15, 15, 28, 28, 15, 15, 15, 15, 15, 15, 29, 29, 30, 30, 29, 29, 15, 15, 15, 15 };
						return center[step];
					}

					return 31;
				}
			}
  
			public void Animate(Level level, int x, int y)
			{
				if (this.isAlive) 
				{
					switch (this.direction)
					{
						case Directions.Left:
							this.step++;
							if (this.step >= 6) 
							{
								this.step = 0;
							}
							break;
	  
						case Directions.Right:
							this.step++;
							if (this.step >= 6) 
							{
								this.step = 0;
							}
							break;
	  
						case Directions.Up:
							this.step++;
							if (this.step >= 2) 
							{
								this.step = 0;
							}
							break;
	  
						case Directions.Down:
							this.step++;
							if (this.step >= 2) 
							{
								this.step = 0;
							}
							break;
	  
						case Directions.None:
							this.step++;
							if (this.step >= 30) 
							{
								this.step = 0;
							}
							break;
					}
				}
			}
  
			public void Move(Level level)
			{
				if (this.isAlive)
				{
					Position p = level.GetSpritePosition(this);
					Position d = new Position(p);
					Position z = new Position(d);
	  
					direction = Directions.None;
					if (Keys[0])
					{
						z.X--; direction = Directions.Left;
					} 
					else
					{
						stone[0] = false;
						if (Keys[1])
						{
							z.X++; direction = Directions.Right;
						}
						else
						{
							stone[1] = false;
							if (Keys[2])
							{
								z.Y--; direction = Directions.Up;
							}
							else 
							{
								if (Keys[3])
								{ 
									z.Y++; direction = Directions.Down;
								}
							}
						} 
					}
	  
					if (d != z)
					{
						// Nothing
						if (level.GetSprite(z.X, z.Y).GetType() == typeof(Nothing))
						{
							level.SetSprite(d.X, d.Y, this);
						}
	  
						// Diamond
						if (level.GetSprite(z.X, z.Y).GetType() == typeof(Diamond))
						{
							level.Collected += 1;
							level.Score += 3;
						}
	  
						// Stone
						if (level.GetSprite(z.X, z.Y).GetType() == typeof(Stone))
						{
							if ((z.X > d.X) && (level.GetSprite(z.X + 1, z.Y).GetType() == typeof(Nothing)))
							{
								if (stone[1])
								{
									level.SetSprite(d.X + 2, d.Y, level.GetSprite(d.X + 1, d.Y));
									level.SetSprite(d.X + 1, d.Y, new Nothing());
								}
								stone[1] = !stone[1];
							}
	  
							if ((z.X < d.X) && (level.GetSprite(z.X - 1, z.Y).GetType() == typeof(Nothing)))
							{
								if (stone[0])
								{
									level.SetSprite(d.X - 2, d.Y, level.GetSprite(d.X - 1, d.Y));
									level.SetSprite(d.X - 1, d.Y, new Nothing());
								}
								stone[0] = !stone[0];
							}
						}
	  
						// Ground
						if (level.GetSprite(z.X, z.Y).GetType() == typeof(Nothing) || level.GetSprite(z.X, z.Y).GetType() == typeof(Ground) || level.GetSprite(z.X, z.Y).GetType() == typeof(Diamond))
						{
							level.SetSprite(z.X, z.Y, this);
							level.SetSprite(d.X, d.Y, new Buffer());
						}
	  
						// Exit
						if (level.GetSprite(z.X, z.Y).GetType() == typeof(Exit)) level.Exit();
	  
						// Ghost
						if (level.GetSprite(z.X, z.Y).GetType().BaseType == typeof(Ghost)) level.Digger.Die();
					}
	  
					this.Animate(level, z.X, z.Y);
				}
			}
  
			public void Die()
			{
				this.isAlive = false;
			}
		}
  
		private class Exit : Sprite 
		{
			public override byte ImageIndex
			{ 
				get
				{
					return 32; 
				}
			}
		}
  
		private class Changer : Sprite
		{
			public override byte ImageIndex
			{
				get
				{ 
					return 33; 
				}
			}
		}
  
		private class Gravitation : Sprite
		{
			public void Simulate(Level l, int x, int y)
			{
				int dx = x;
				int dy = y;
  
				// Nothing
				if (l.GetSprite(x, y + 1).GetType() == typeof(Nothing))
				{
					dy = y + 1;
				}
				else
				{
					// Stone or diamond
					if ((l.GetSprite(x, y + 1).GetType() == typeof(Stone)) || (l.GetSprite(x, y + 1).GetType() == typeof(Diamond)))
					{
						if (l.GetSprite(x - 1, y + 1).GetType() == typeof(Nothing) && l.GetSprite(x - 1, y).GetType() == typeof(Nothing))
						{
							dx = x - 1;
							dy = y + 1;
						}
						else
						{
							if (l.GetSprite(x + 1, y + 1).GetType() == typeof(Nothing) && l.GetSprite(x + 1, y).GetType() == typeof(Nothing))
							{
								dx = x + 1;
								dy = y + 1;
							}
						}
					}
  
					// Changer
					if ((l.GetSprite(x, y + 1).GetType() == typeof(Changer)) && (l.GetSprite(x, y).GetType() == typeof(Stone)) && (l.GetSprite(x, y + 2).GetType() == typeof(Nothing)))
						dy = y + 2;
				}
  
				// Mark
				if ((dx != x) || (dy != y)) l.SetSprite(dx, dy, new Marker());
			}
  
			public void Move(Level l, int x, int y)
			{
				int dx = x;
				int dy = y;
  
				// Follow gravitation
				if (l.GetSprite(x, y + 1).GetType() == typeof(Marker))
				{
					dy = y + 1;
				}
				else
				{
					// Falls to the left or to the right
					if (l.GetSprite(x, y + 1).GetType() == typeof(Stone) || l.GetSprite(x, y + 1).GetType() == typeof(Diamond) || l.GetSprite(x, y + 1).GetType() == typeof(Nothing))
					{
						if (l.GetSprite(x - 1, y + 1).GetType() == typeof(Marker) && (l.GetSprite(x - 1, y).GetType() == typeof(Nothing) || l.GetSprite(x - 1, y).GetType() == typeof(Marker)))
						{
							dx = x - 1;
							dy = y + 1;
						}
						else
						{
							if (l.GetSprite(x + 1, y + 1).GetType() == typeof(Marker) && (l.GetSprite(x + 1, y).GetType() == typeof(Nothing) || l.GetSprite(x + 1, y).GetType() == typeof(Marker)))
							{
								dx = x + 1;
								dy = y + 1;
							}
						}
					}
  
					// Changer
					if (l.GetSprite(x, y + 1).GetType() == typeof(Changer) && l.GetSprite(x, y).GetType() == typeof(Stone) && l.GetSprite(x, y + 2).GetType() == typeof(Marker)) dy = y + 2;
				}
  
				// Move
				if ((dx != x) || (dy != y))
				{
					if ((dy - y) == 2)
					{
						// changer
						l.SetSprite(dx, dy, new Diamond());
					}
					else
					{
						l.SetSprite(dx, dy, l.GetSprite(x, y));
						if (l.GetSprite(dx, dy).GetType() == typeof(UvStone))
							l.SetSprite(dx, dy, new Stone());
					}
  
					l.SetSprite(x, y, new Nothing());      
  
					// Kill digger if necessary
					if (l.GetSprite(dx, dy + 1).GetType() == typeof(Digger)) l.Digger.Die();
  
					if (l.GetSprite(dx, dy + 1).GetType().BaseType == typeof(Ghost))
					{
						Ghost g = (Ghost) l.GetSprite(dx, dy + 1);
						g.Blast(l);
					}
				}
			}
		}
  
		private class Marker : Gravitation
		{
			public override byte ImageIndex
			{ 
				get
				{
					return 31; 
				}
			}
		}
  
		private class Stone : Gravitation
		{
			public override byte ImageIndex
			{ 
				get
				{
					return 1; 
				}
			}
		}
  
		private class Diamond : Gravitation
		{
			private static byte blinkCounter = 0;
  
			public static void Blink()
			{
				blinkCounter++;
				if (blinkCounter == 6) 
				{
					blinkCounter = 0;	
				}
			}
  
			public override byte ImageIndex
			{
				get
				{
					return (byte) (13 - ((blinkCounter + 4) % 6));
				}
			}
		}
  
		private class Nothing : Sprite 
		{ 
		}
		
		private class Buffer : Sprite 
		{
		}
		
		private class UvStone : Gravitation
		{ 
		}
  
		private class Level
		{
			private byte[] data;
			private Sprite[][] map = null;
			private Ghost[] ghosts = null;
			public Digger Digger;
			public int Number;
			public int Diamonds;
			public int Collected;
			public int Time;
			public int Score = 0;
  
			public Level()
			{
				using (Stream stream = GetType().Assembly.GetManifestResourceStream("Digger.Level.bin"))
				{
					this.data = new Byte[stream.Length];
					stream.Read(this.data, 0, (int) stream.Length);
				}
			}
  
			public void Load()
			{
				this.ghosts = new Ghost[16];
				
				this.map = new Sprite[20][];
				for (int i = 0; i < 20; i++)
				{
					this.map[i] = new Sprite[14];
				}
				
				byte[] level = new byte[156];
				for (int i = 0; i < 156; i++)
				{
					level[i] = this.data[this.Number * 156 + i];
				}
      
				for (int y = 0; y < 14; y++)
				{
					for (int x = 0; x < 10; x++)
					{
						this.CreateSprite(x * 2, y , (byte) (level[y * 10 + x] >> 4));
						this.CreateSprite(x * 2 + 1, y , (byte) (level[y * 10 + x] & 0x0F));
					}
				}
  
				Digger = (Digger) this.map[level[145]][level[146] - 2];
  
				for (int g = 0; g < 16; g++) 
				{
					ghosts[g] = null;
				}

				int Ghost = 0;
				for (int y = 0; y < 14; y++)
				{
					for (int x = 0; x < 20; x++)
					{
						if (GetSprite(x, y).GetType().BaseType == typeof(Ghost))
						{
							Ghost g = (Ghost) GetSprite(x, y);
							ghosts[Ghost] = g;
							int GhostInfo = level[0x94 + (Ghost >> 1)];
							if ((Ghost & 1) != 0)
							{
								GhostInfo = GhostInfo & 0x0F;
								if (g.GetType() == typeof(Ghost90LR)) 
								{
									((Ghost90LR) g).Lastturn = Directions.Right;
								}
							}
							else
							{
								GhostInfo = GhostInfo >> 4;
								if (g.GetType() == typeof(Ghost90LR)) 
								{
									((Ghost90LR) g).Lastturn = Directions.Left;
								}
							}
					
							Directions[] directions = { Directions.Down, Directions.Up, Directions.Right, Directions.Left };
							g.Direction = directions[GhostInfo];
							Ghost++;
						}
					}
				}
  
				this.Diamonds = (level[147] / 0x10) * 10 + (level[147] % 0x10);
				this.Collected = 0;
				this.Time = 5000;
			}
  
			public void Exit()
			{ 
				if (this.Collected < this.Diamonds) 
				{
					return;
				}
				
				if (this.Number >= 29) 
				{
					return;
				}
				
				this.Number++;
				this.Load();
			}
  
			public void CreateSprite(int x, int y, int id)
			{
				Type[] type = { typeof(Nothing), typeof(Stone), typeof(Ground), typeof(Ghost180), null, typeof(Diamond), typeof(Wall), typeof(Ghost90L), null, typeof(UvStone), typeof(Digger), typeof(Ghost90LR), typeof(Exit), null, typeof(Changer), typeof(Ghost90R) };
				this.SetSprite(x, y, (Sprite) Activator.CreateInstance(type[id]));
			} 
  
			public void SetSprite(int x, int y, Sprite sprite)
			{
				this.map[x][y] = sprite;
			}
  
			public Sprite GetSprite(int x, int y)
			{
				return this.map[x][y];
			}
  
			public Position GetSpritePosition(Sprite s)
			{
				for (int x = 0; x < 20; x++)
				{
					for (int y = 0; y < 14; y++)
					{
						if (this.GetSprite(x, y) == s)
						{
							return new Position(x, y);
						}
					}
				}

				return new Position();
			}
  
			public void Tick1()
			{
				Diamond.Blink();
			}
  
			public void Tick4()
			{
				// Turn buffers into nothing
				for (int y = 13; y >= 0; y--)
				{
					for (int x = 19; x >= 0; x--)
					{
						if (map[x][y].GetType() == typeof(Buffer))
						{
							this.SetSprite(x, y, new Nothing());
						}
					}
				}
  
				// Digger
				Digger.Move(this);
  
				// Stones and Diamonds (UVStones?)
				for (int y = 13; y >= 0; y--)
				{
					for (int x = 19; x >= 0; x--)
					{
						if (this.map[x][y].GetType() == typeof(Stone) || this.map[x][y].GetType() == typeof(Diamond) || this.map[x][y].GetType() == typeof(UvStone))
						{
							Gravitation gravitation = (Gravitation) this.map[x][y];
							gravitation.Simulate(this, x, y);
						}
					}
				}
  
				for (int y = 13; y >= 0; y--)
				{
					for (int x = 19; x >= 0; x--)
					{
						if (this.map[x][y].GetType() == typeof(Stone) || this.map[x][y].GetType() == typeof(Diamond) || this.map[x][y].GetType() == typeof(UvStone))
						{
							Gravitation gravitation = (Gravitation) this.map[x][y];
							gravitation.Move(this, x, y);
						}
					}
				}
  
				for (int i = 0; i < 16; i++)
				{
					if (ghosts[i] != null) 
					{
						ghosts[i].Move(this);
					}
				}

				if (this.Time > 0) 
				{
					this.Time--;
				}

				if (this.Time == 0) 
				{
					Digger.Die();
				}
			}
		}
	}
}
